<TeXmacs|2.1.2>

<style|source>

<\body>
  <active*|<\src-title>
    <src-package-dtd|steelblue-combo|1.0|steelblue-combo|1.0>

    <\src-purpose>
      Steelblue theme for presentations and posters.
    </src-purpose>

    <src-copyright|2013--2023|Joris van der Hoeven>

    <\src-license>
      This software falls under the <hlink|GNU general public license,
      version 3 or later|$TEXMACS_PATH/LICENSE>. It comes WITHOUT ANY
      WARRANTY WHATSOEVER. You should have received a copy of the license
      which the software. If not, see <hlink|http://www.gnu.org/licenses/gpl-3.0.html|http://www.gnu.org/licenses/gpl-3.0.html>.
    </src-license>
  </src-title>>

  <use-package|bluish-combo|steelblue-deco>

  <copy-theme|steel-blue|bluish>

  <select-theme|steel-blue|steelblue>

  <select-theme|steel-blue|steelblue-bright-scene>

  <\active*>
    <\src-comment>
      Titles
    </src-comment>
  </active*>

  <select-title-deco|steel-blue|steelblue>

  <assign-uniform|steel-blue-title|#d8c0c0>

  <\active*>
    <\src-comment>
      Sessions
    </src-comment>
  </active*>

  <select-input-deco|steel-blue|steelblue-light>

  <select-fold-title-deco|steel-blue|steelblue-dark>

  <select-fold-bar-deco|steel-blue|steelblue-dark>

  <\active*>
    <\src-comment>
      Posters
    </src-comment>
  </active*>

  <assign|steel-blue-title-block|<value|steelblue-block>>

  <assign|steel-blue-framed-block|<value|steelblue-block>>

  <assign|steel-blue-framed-block*|<value|steelblue-titled-block>>
</body>

<\initial>
  <\collection>
    <associate|sfactor|7>
  </collection>
</initial>