
/******************************************************************************
* MODULE     : cairo_renderer.hpp
* DESCRIPTION: Cairo drawing interface class
* COPYRIGHT  : (C) 2008 Massimiliano Gubinelli
*******************************************************************************
* This software falls under the GNU general public license version 3 or later.
* It comes WITHOUT ANY WARRANTY WHATSOEVER. For details, see the file LICENSE
* in the root directory or <http://www.gnu.org/licenses/gpl-3.0.html>.
******************************************************************************/

#ifndef CAIRO_RENDERER_HPP
#define CAIRO_RENDERER_HPP
#include "basic_renderer.hpp"

basic_renderer_rep* the_cairo_renderer();

#endif // defined CAIRO_RENDERER_HPP
